
-- # import prelude
-- # import id
-- # import prop
-- # import eqchain

data PartialBijection (A, B : Set) : Type where
  pBij : (mem  : A -> B -> Set) ->
	 (injL : (x,y : A) -> (z : B) -> mem x z -> mem y z -> x == y ) ->
	 (injR : (x : A) -> (y,z : B) -> mem x y -> mem x z -> y == z ) ->
	 PartialBijection A B

∅ (A, B : Set) : PartialBijection A B
∅ = pBij (\_ _ -> False)
	 (\x y z ⊥ -> absElim ⊥ )
	 (\x y z ⊥ -> absElim ⊥ )

infix 15 ∈, ⊆, ≡

(∈) (A, B : Set) : A × B -> PartialBijection A B -> Set
p ∈ pBij mem _ _ = mem (fst p) (snd p)

(⊆) (A,B : Set) : PartialBijection A B -> PartialBijection A B -> Set
Φ ⊆ Ψ = (z : _ × _) -> z ∈ Φ -> z ∈ Ψ

(≡) (A,B : Set) : PartialBijection A B -> PartialBijection A B -> Set
Φ ≡ Ψ = Φ ⊆ Ψ /\ Ψ ⊆ Φ

identity (A : Set) : PartialBijection A A
identity = pBij (==)
		(\x y z xz yz -> trans z xz (sym yz))
		(\x y z xy xz -> trans x (sym xy) xz)

-- When extending we can take stuff away.  Extends Ψ Φ z doesn't mean that
-- Φ ⊆ Ψ.  This is to allow z to override mappings in Φ.
Extends (A,B : Set) : PartialBijection A B -> PartialBijection A B -> A × B -> Set
Extends {A} {B} Ψ Φ p =
  (q : A × B) -> p ∈ Ψ /\ (q ∈ Ψ -> q == p \/ q ∈ Φ)

idExtends (A : Set)(x : A) : Extends identity identity (pair x x)
idExtends z = andI refl orI2

inverse (A,B : Set) : PartialBijection A B -> PartialBijection B A
inverse (pBij mem injL injR) =
    pBij (flip mem)
	 (\x y z xz yz -> injR z x y xz yz)
	 (\x y z xy xz -> injL y z x xy xz)

inverseInverseId (A,B : Set)(Φ : PartialBijection A B) :
    inverse (inverse Φ) ≡ Φ
inverseInverseId {_} {_} {pBij mem _ _} = andI (\z el -> el) (\z el -> el)

inverseIdentity (A : Set) : inverse identity ≡ identity {A}
inverseIdentity = andI (\x -> sym) (\x -> sym)

elemInverse (A,B : Set)(Φ : PartialBijection A B)(p : A × B) :
    p ∈ Φ -> swap p ∈ inverse Φ
elemInverse {_} {_} {pBij mem _ _} el = el

elemInverse' (A,B : Set)(Φ : PartialBijection A B)(p : B × A) :
    p ∈ inverse Φ -> swap p ∈ Φ
elemInverse' {_} {_} {pBij mem _ _} el = el

extInverse (A,B : Set)(Φ, Ψ : PartialBijection A B)(p : A × B) :
    Extends Ψ Φ p -> Extends (inverse Ψ) (inverse Φ) (swap p)
extInverse {A} {B} {Φ} {Ψ} {pair x y} ext (pair z w) =
    andI (elemInverse $ andE1 lem)
	 (\el -> let el' : pair w z ∈ Ψ
		     el' = elemInverse' {_} {_} {_} {pair z w} el
	         in orE (andE2 lem el')
			(orI1 ∘ cong swap)
			(orI2 ∘ elemInverse)
	 )
  where
    lem = ext (pair w z)

extEqual1 (A,B : Set)(Φ, Ψ1, Ψ2 : PartialBijection A B)(p : A × B) :
    Ψ1 ≡ Ψ2 -> Extends Ψ1 Φ p -> Extends Ψ2 Φ p
extEqual1 {_} {_} {Φ} {Ψ1} {Ψ2} {p} eq ext q =
    andI (rem1 p rem3)
	 (\qinΨ2 -> rem4 $ rem2 q qinΨ2)
  where
    rem1 : Ψ1 ⊆ Ψ2
    rem1 = andE1 eq

    rem2 : Ψ2 ⊆ Ψ1
    rem2 = andE2 eq

    rem3 : p ∈ Ψ1
    rem3 = andE1 (ext q)

    rem4 : q ∈ Ψ1 -> q == p \/ q ∈ Φ
    rem4 = andE2 (ext q)

extEqual2 (A,B : Set)(Φ1, Φ2, Ψ : PartialBijection A B)(p : A × B) :
    Φ1 ≡ Φ2 -> Extends Ψ Φ1 p -> Extends Ψ Φ2 p
extEqual2 {_} {_} {Φ1} {Φ2} {Ψ} {p} eq ext q =
    andI (andE1 (ext q))
	 (\qinΨ -> orE (andE2 (ext q) qinΨ) orI1 (orI2 ∘ rem1 q))
  where
    rem1 : Φ1 ⊆ Φ2
    rem1 = andE1 eq

extIdInverse (A : Set)(Φ : PartialBijection A A)(p : A × A) :
    Extends Φ identity p -> Extends (inverse Φ) identity (swap p)
extIdInverse {A} {Φ} {pair x y} ext =
    extEqual2 {_} {_} {inverse identity} {identity} inverseIdentity rem1
  where
    rem1 : Extends (inverse Φ) (inverse identity) (pair y x)
    rem1 = extInverse {_} {_} {identity} ext

