
\section{The QuickCheck Library} \label{QuickCheck.lagda}

\begin{code}
-- # import Property
-- # import Monad
-- # import Nat
-- # import List
-- # import Bool
\end{code}

The QuickCheck library defines functions to build the generators used to
generate the test cases, and provides special property constants to 

%- Generators -------------------------------------------------------------
\subsection{Building Test Data Generators}

\begin{code}

Gen : (a : Set) -> Set

-- CanGenerate : (A : Set) -> Gen A -> A -> Prop
-- CanGenerate A g x = { size : Nat; CanProduce A (g size) x }

-- primitives
returnGen (a	: Set) : a -> Gen a
bindGen	  (a, b : Set) : Gen a -> (a -> Gen b) -> Gen b

monadGen : Monad Gen
monadGen =
  struct return = returnGen
	 bind	= bindGen

sized	  (a	: Set) : (Nat -> Gen a) -> Gen a
promote	  (a, b : Set) : (a -> Gen b) -> Gen (a -> b)
variant	  (a	: Set) : Nat -> Gen a -> Gen a
elements  (a	: Set) : List a -> Gen a
oneof	  (a	: Set) : List (Gen a) -> Gen a
frequency (a	: Set) : List {Nat; Gen a} -> Gen a

\end{code}

%- Standard generators ----------------------------------------------------
\subsection{Some Standard Generators}

\begin{code}
-- Independent functions
arbFun (a, b : Set) : (a -> Gen b -> Gen b) -> Gen b -> Gen (a -> b)
arbFun coarb genb = promote (\x -> coarb x genb)

-- Sigma types
arbSigma (a : Set)(b : a -> Set) : Gen a -> ((x : a) -> Gen (b x)) -> Gen {x : a; b x}
arbSigma gA fB =
  monadGen.do x <- gA
	      y <- fB x
	      monadGen.return (x,y)

-- Natural numbers
arbNat : Gen Nat
arbNat = sized arbNatS
  where
    arbNatS : Nat -> Gen Nat
    arbNatS 0        = monadGen.return 0
    arbNatS (suc n)  =
      frequency [ (1, monadGen.return 0)
		, (5, liftM monadGen suc (arbNatS n))
		]

coarbNat (a : Set) : Nat -> Gen a -> Gen a
coarbNat n g = variant n g

-- Booleans
arbBool : Gen Bool
arbBool = elements [false, true]

coarbBool (a : Set) : Bool -> Gen a -> Gen a
coarbBool false g = variant 0 g
coarbBool true  g = variant 1 g

-- Lists
arbList (a : Set) : Gen a -> Gen (List a)
arbList g = sized (arbListS g)
  where
    arbListS (a : Set) : Gen a -> Nat -> Gen (List a)
    arbListS genA 0       = monadGen.return []
    arbListS genA (suc n) =
      frequency [ (1, monadGen.return [])
		, (5, liftM2 monadGen (::) genA (arbListS genA n))
		]

\end{code}

%- Property constants -----------------------------------------------------
\subsection{Special Property Constants}

\begin{code}
Tested : Prop -> Prop

forAll	 (a : Set) : Gen a -> (a -> Prop) -> Prop
collect  (a : Set) : a -> Prop -> Prop
label	 (a : Set) : a -> Prop -> Prop
classify (a : Set) : Bool -> a -> Prop -> Prop
trivial   	   : Bool -> Prop -> Prop
\end{code}

