
module Plugins.Rewrite where

import Control.Monad.Trans

import Syntax.Position
import Syntax.Internal
import Syntax.Values
import Syntax.Scope

import TypeChecker.Monad
import TypeChecker.Monad.Errors
import TypeChecker.Monad.Options
import TypeChecker.Errors
import TypeChecker.Abstract
import TypeChecker.UnfoldConstants
import TypeChecker.ValueToTerm
import TypeChecker.Abstractions
import TypeChecker

import Utils.Error

import Plugins

plugin :: SomePlugin
plugin = mkPlugin $ 
    Plugin  { plName		= "rewrite"
            , plHelp            = help
	    , plInit		= initialize
	    , plFinal		= finalize
	    , plCheck		= check
	    , plInfer		= infer
	    , plOptions		= options
	    , plDefaultOptions	= defaultOptions
	    }

type State = ()
type Options = ()

help = "Rewrite tactics: rewrite-plugin(eq,t) uses eq to rewrite the goal (t is the proof of the new goal)."

initialize :: Options -> TC State
initialize _ = return ()

finalize :: Options -> State -> TC ()
finalize _ _ = return ()

check :: Options -> State -> Range -> [Term] -> Type -> TC (Value, State)
check o s r (eq:es@(_:_:_)) t =
    do	scope <- getScopeEnv
	v <- checkType ( CallPlugin r scope "rewrite"
			 [eq, CallPlugin r scope "rewrite" es]
		       ) t
	return (v,s)
check _ s _ [eq,rest] t =
    do	(_,teq) <- inferType eq
	scope <- getScopeEnv
	let equal = getName scope "=="
	    subst = Def noRange $ getName scope "subst"
	    sym   = Def noRange $ getName scope "sym"
	v <- unfoldConstants $ unEl teq
	case vAppView v of
	    VAVDef x _ [_,w1,w2]
		| x == equal ->
		do  vs <- abstract w1 $ unEl t
		    verbose 15 $ liftIO $ putStrLn $ "rewrite-plugin: [" ++ show w1 ++ "] " ++ show t ++ " = " ++ show (map (flip instantiate [vId (dummyName "HOLE")]) $ init vs)
		    v <- case init vs of
			[]	-> failure $ customError $ "failed to apply rewrite " ++ show v
			v:_	-> return v -- apply rewriting everywhere possible
		    p <- valueToTerm $ vLam $ mkAbs "HOLE" v
		    let e = subst `App` p `App` (sym `App` eq) `App` rest
		    v <- checkType e t
		    verbose 5 $ liftIO $ putStrLn $ "[rewrite-plugin] " ++ show e ++ " : " ++ show t
		    return (v, s)
	    _ -> failure $ customError "the first argument to the rewrite-plugin should be an equality proof"
check _ _ _ _ _ = failure $ customError "the rewrite-plugin expects at least two arguments"

infer :: Options -> State -> Range -> [Term] -> TC (Value, Type, State)
infer _ _ _ _ = failure $ customError "can't infer the type of a rewrite"

options :: PluginOptions Options
options = simpleOptions []

defaultOptions :: Options
defaultOptions = ()

