{-# OPTIONS -fglasgow-exts #-}
module Syntax.Scope where

import Data.Map as Map
import Data.Set as Set
import Data.Generics

import Syntax.Name
import qualified Syntax.Concrete as C
import Syntax.Position

data BoundName = BName  { kindOfName :: Kind
			, actualName :: Name
			}
    deriving (Typeable, Data)

instance Show BoundName where
    show = nameString . actualName


instance HasRange BoundName where
    getRange = getRange . actualName

data Kind = DefName | ConName | DataName | VarName | SigName
    deriving (Eq, Typeable, Data)

instance Show Kind where
    show DefName  = "defined name"
    show ConName  = "constructor name"
    show DataName = "datatype name"
    show VarName  = "variable name"
    show SigName  = "signature name"

data ScopeEnv = ScopeEnv
		{ eNames      :: Map.Map String BoundName
		, eFixities   :: Map.Map String C.Fixity
		, eSignatures :: Map.Map Name (Set Name)    -- fields of the signatures
		}
    deriving (Typeable, Data)

emptyScopeEnv = ScopeEnv Map.empty Map.empty Map.empty

addDefinedName :: Name -> ScopeEnv -> ScopeEnv
addDefinedName x e = e { eNames = addName (BName DefName x) $ eNames e }

addVarNames :: [Name] -> ScopeEnv -> ScopeEnv
addVarNames xs e = e { eNames = foldr (addName . BName VarName) (eNames e) xs }

addName :: BoundName -> Map String BoundName -> Map String BoundName
addName (BName _ DontCare) m = m
addName x m = Map.insert (nameString $ actualName x) x m

getName :: ScopeEnv -> String -> Name
getName scope s =
    case Map.lookup s $ eNames scope of
	Nothing		 -> error $ "Syntax.Scope.getName: no such name " ++ s
	Just (BName _ x) -> x

